from django.urls import path

from huscy.appointments.feed import AppointmentFeed
from huscy.appointments.feed import get_feed_url


urlpatterns = [
    path('feed/<str:token>', AppointmentFeed(), name='feed'),
    path('feedurl/', get_feed_url, name='feed-url'),
]



import secrets

from django.conf import settings
from django.db import models
from django.db.models import CASCADE
from django.db.models import CharField
from django.db.models import DateTimeField
from django.db.models import ForeignKey
from django.db.models import Model
from django.db.models import TextField
from django.utils.translation import gettext_lazy as _


class Appointment(Model):
    owner = ForeignKey(settings.AUTH_USER_MODEL, on_delete=CASCADE, verbose_name=_('Owner'))
    title = CharField(max_length=255, blank=True, default=_('New appointment'),
                      verbose_name=_('Title'))
    description = TextField(blank=True, default='', verbose_name=_('Description'))
    start = DateTimeField(verbose_name=_('Start'))
    end = DateTimeField(verbose_name=_('End'))
    resource = CharField(max_length=128, blank=True, default='', verbose_name=_('Resource'))

    def __str__(self):
        return str(self.title)

    class Meta:
        verbose_name = _('Appointment')
        verbose_name_plural = _('Appointments')


class Invitation(models.Model):
    class STATUS(models.IntegerChoices):
        pending = 0, _('Pending')
        accepted = 1, _('Accepted')
        declined = 2, _('Declined')

    appointment = models.ForeignKey(Appointment, on_delete=CASCADE, related_name='invitations',
                                    verbose_name=_('Appointment'))
    participant = models.CharField(_('Participant'), max_length=128)
    status = models.PositiveSmallIntegerField(_('Status'), choices=STATUS.choices,
                                              default=STATUS.pending)

    def __str__(self):
        return f'{self.appointment.title} -- {self.participant} ({self.get_status_display()})'

    class Meta:
        verbose_name = _('Invitation')
        verbose_name_plural = _('Invitations')


class Reminder(Model):
    appointment = ForeignKey(Appointment, on_delete=CASCADE, related_name='reminders',
                             verbose_name=_('Appointment'))
    remind_at = DateTimeField(verbose_name=_('Remind at'))

    class Meta:
        verbose_name = _('Reminder')
        verbose_name_plural = _('Reminders')


class Resource(Model):
    name = CharField(max_length=128, verbose_name=_('Name'))

    def __str__(self):
        return str(self.name)

    class Meta:
        verbose_name = _('Resource')
        verbose_name_plural = _('Resources')


def create_token():
    return secrets.token_urlsafe(96)


class Token(Model):
    user = ForeignKey(settings.AUTH_USER_MODEL, on_delete=CASCADE, verbose_name=_('User'))
    key = CharField(max_length=128, unique=True, default=create_token, verbose_name=_('Key'))

    class Meta:
        verbose_name = _('Token')
        verbose_name_plural = _('Tokens')



from django.db import migrations, models
import huscy.appointments.models


class Migration(migrations.Migration):

    dependencies = [
        ('appointments', '0002_auto_20200226_0735'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='key',
            field=models.CharField(default=huscy.appointments.models.create_token, max_length=128, unique=True, verbose_name='Key'),
        ),
    ]



from django.contrib import admin

from huscy.appointments import models

admin.site.register(models.Appointment)
admin.site.register(models.Invitation)
admin.site.register(models.Reminder)
admin.site.register(models.Resource)



import warnings
from datetime import datetime

from django.db import transaction

from huscy.appointments import models, validators


@transaction.atomic
def create_appointment(creator, start, end, title='', description='', resource=None,
                       participants=[]):
    validators.validate_start_end(start, end)
    validators.validate_resource_exists(resource)

    if resource:
        appointments = get_appointments(from_time=start, to_time=end, resource=resource)
        if appointments.exists():
            raise ValueError("Resource is already in use within this time.")

    appointment = models.Appointment.objects.create(
        owner=creator,
        start=start,
        end=end,
        title=title,
        description=description,
        resource=resource or '',
    )

    if participants:
        set_participants(appointment, participants)

    return appointment


@transaction.atomic
def update_appointment(appointment, owner=None, title=None, description=None, resource=None,
                       start=None, end=None, participants=None):

    if start is not None and end is not None:
        validators.validate_start_end(start, end)
        appointment.start = start
        appointment.end = end
    elif start is not None:
        validators.validate_start_end(start, appointment.end)
        appointment.start = start
    elif end is not None:
        validators.validate_start_end(appointment.start, end)
        appointment.end = end

    if resource is not None:
        validators.validate_resource_exists(resource)
        appointment.resource = resource

    if resource:
        appointments = get_appointments(from_time=appointment.start, to_time=appointment.end,
                                        resource=appointment.resource)
        if appointments.exclude(pk=appointment.pk).exists():
            raise ValueError("Resource is already in use within this time.")

    appointment.owner = owner or appointment.owner
    appointment.title = title or appointment.title
    appointment.description = description or appointment.description

    appointment.save()

    # now proceed to set participants
    if participants is not None:
        set_participants(appointment, participants)

    return appointment


def remove_appointment(appointment):
    appointment.delete()


def get_appointments(from_time=datetime.min, to_time=datetime.max, resource=None, owner=None):
    qs = models.Appointment.objects
    qs = qs.exclude(start__gte=to_time)
    qs = qs.exclude(end__lte=from_time)
    if resource:
        if type(resource) == list:
            qs = qs.filter(resource__in=resource)
        else:
            qs = qs.filter(resource=resource)
    if owner:
        qs = qs.filter(owner=owner)
    qs = qs.order_by('-start', 'end', 'resource')
    return qs


def get_participants(appointment):
    return dict(appointment.invitations.values_list('participant', 'status'))


def set_participants(appointment, participants):
    current_participants = set(appointment.invitations.values_list('participant', flat=True))

    # create new participants
    new_participants = set(participants) - set(current_participants)
    new_invitations = [models.Invitation(appointment=appointment, participant=participant)
                       for participant in new_participants]
    models.Invitation.objects.bulk_create(new_invitations)

    # remove participants
    qs = models.Invitation.objects.filter(appointment=appointment,
                                          participant__in=(
                                              set(current_participants)-set(participants)
                                          ))
    qs.delete()

    return appointment


def accept_appointment(appointment, participant):
    if isinstance(appointment, models.Appointment):
        invitation = appointment.invitations.get(participant=participant)
    else:
        invitation = (models.Invitation.objects
                                       .select_related('appointment')
                                       .get(appointment_id=appointment, participant=participant))
        warnings.warn('Passing the appointment id to accept_appointment is marked as deprecated. '
                      'Please use appointment object instead.', DeprecationWarning)

    invitation.status = models.Invitation.STATUS.accepted
    invitation.save()
    return invitation.appointment


def decline_appointment(appointment, participant):
    if isinstance(appointment, models.Appointment):
        invitation = appointment.invitations.get(participant=participant)
    else:
        invitation = (models.Invitation.objects
                                       .select_related('appointment')
                                       .get(appointment_id=appointment, participant=participant))
        warnings.warn('Passing the appointment id to decline_appointment is marked as deprecated. '
                      'Please use appointment object instead.', DeprecationWarning)

    invitation.status = models.Invitation.STATUS.declined
    invitation.save()
    return invitation.appointment


def get_resources():
    return models.Resource.objects.all()



from rest_framework import mixins
from rest_framework import permissions
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from huscy.appointments import filters, serializers, services


class AppointmentsViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin, mixins.ListModelMixin,
                          mixins.UpdateModelMixin, viewsets.GenericViewSet):
    filterset_class = filters.AppointmentFilter
    permission_classes = (permissions.IsAuthenticated, )
    serializer_class = serializers.AppointmentSerializer

    def get_queryset(self):
        user = self.request.user
        if user.is_superuser:
            return services.get_appointments()
        return services.get_appointments(owner=user)

    @action(detail=True, methods=['GET'], name='participants')
    def participants(self, request, *args, **kwargs):
        participants = services.get_participants(self.get_object())
        return Response(participants)

    @action(detail=True, methods=['POST'], name='accept appointment')
    def accept(self, request, *args, **kwargs):
        appointment = self.get_object()
        participant = request.data['participant']
        if appointment and participant:
            app = services.accept_appointment(appointment, participant)
            return Response(serializers.AppointmentSerializer(app).data)
        return Response("Something was going wrong!")

    @action(detail=True, methods=['POST'], name='decline appointment')
    def decline(self, request, *args, **kwargs):
        appointment = self.get_object()
        participant = request.data['participant']
        if appointment and participant:
            app = services.decline_appointment(appointment, participant)
            return Response(serializers.AppointmentSerializer(app).data)
        return Response("Something was going wrong!")

    def perform_destroy(self, instance):
        services.remove_appointment(instance)


class ResourcesViewSet(viewsets.ModelViewSet):
    permission_classes = (permissions.IsAuthenticated, )
    queryset = services.get_resources()
    serializer_class = serializers.ResourceSerializer



from rest_framework.routers import DefaultRouter

from huscy.appointments import views


router = DefaultRouter()
router.register('appointments', views.AppointmentsViewSet, basename='appointment')
router.register('resources', views.ResourcesViewSet)

urlpatterns = router.urls



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import huscy.appointments.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('appointments', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='appointment',
            options={'verbose_name': 'Appointment', 'verbose_name_plural': 'Appointments'},
        ),
        migrations.AlterModelOptions(
            name='invitation',
            options={'verbose_name': 'Invitation', 'verbose_name_plural': 'Invitations'},
        ),
        migrations.AlterModelOptions(
            name='reminder',
            options={'verbose_name': 'Reminder', 'verbose_name_plural': 'Reminders'},
        ),
        migrations.AlterModelOptions(
            name='resource',
            options={'verbose_name': 'Resource', 'verbose_name_plural': 'Resources'},
        ),
        migrations.AlterModelOptions(
            name='token',
            options={'verbose_name': 'Token', 'verbose_name_plural': 'Tokens'},
        ),
        migrations.AlterField(
            model_name='appointment',
            name='description',
            field=models.TextField(blank=True, default='', verbose_name='Description'),
        ),
        migrations.AlterField(
            model_name='appointment',
            name='end',
            field=models.DateTimeField(verbose_name='End'),
        ),
        migrations.AlterField(
            model_name='appointment',
            name='owner',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='Owner'),
        ),
        migrations.AlterField(
            model_name='appointment',
            name='resource',
            field=models.CharField(blank=True, default='', max_length=128, verbose_name='Resource'),
        ),
        migrations.AlterField(
            model_name='appointment',
            name='start',
            field=models.DateTimeField(verbose_name='Start'),
        ),
        migrations.AlterField(
            model_name='appointment',
            name='title',
            field=models.CharField(blank=True, default='New appointment', max_length=255, verbose_name='Title'),
        ),
        migrations.AlterField(
            model_name='invitation',
            name='appointment',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='invitations', to='appointments.Appointment', verbose_name='Appointment'),
        ),
        migrations.AlterField(
            model_name='invitation',
            name='participant',
            field=models.CharField(max_length=128, verbose_name='Participant'),
        ),
        migrations.AlterField(
            model_name='invitation',
            name='status',
            field=models.PositiveSmallIntegerField(choices=[(0, 'Pending'), (1, 'Accepted'), (2, 'Declined')], default=0, verbose_name='Status'),
        ),
        migrations.AlterField(
            model_name='reminder',
            name='appointment',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reminders', to='appointments.Appointment', verbose_name='Appointment'),
        ),
        migrations.AlterField(
            model_name='reminder',
            name='remind_at',
            field=models.DateTimeField(verbose_name='Remind at'),
        ),
        migrations.AlterField(
            model_name='resource',
            name='name',
            field=models.CharField(max_length=128, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='token',
            name='key',
            field=models.CharField(default=huscy.appointments.models.create_token, max_length=64, unique=True, verbose_name='Key'),
        ),
        migrations.AlterField(
            model_name='token',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
    ]



from huscy.appointments.models import Resource


def validate_start_end(start, end):
    if start > end:
        raise ValueError("End must be greater then start.")
    elif start == end:
        raise ValueError("Start and end are the same.")


def validate_resource_exists(resource):
    if resource and not Resource.objects.filter(name=resource).exists():
        raise ValueError(f'Resource: {resource} does not exist')
